import os
from skitai.server.wastuff import process, daemon    
from aquests.lib import pathtool

DEFAULT = """[smtpda]
# SMTP Delivery Agent
max-retry = 10
keep-days = 1
smtp-server =
user =
password =
ssl = true

[:crontab]
# Example: 0/5 * * * * /usr/bin/python3 something
"""

_default_conf = os.path.join (os.environ ["HOME"], ".skitai.conf")
if not os.path.exists (_default_conf):
    with open (_default_conf + ".example", "w") as f:
        f.write (DEFAULT)    
    _default_log_dir = daemon.get_default_logpath ()
    _default_var_dir = daemon.get_default_varpath ()
    
else:
    _default_log_dir = "/var/log/skitai/default"
    _default_var_dir = "/var/tmp/skitai/default"

pathtool.mkdir (_default_log_dir)
pathtool.mkdir (_default_var_dir)



#------------------------------------------------------------
# Cookie and Session Adaptor Between Skitai and Django 
# 
# Available:
#
#   request.user
#   request.session
#
# Jan 10, 2018
# Hans Roh
#------------------------------------------------------------

try:
    from django.contrib import auth
    from django.core.handlers.wsgi import WSGIRequest
    from django.utils.functional import SimpleLazyObject
    from django.contrib.auth.middleware import get_user
    from django.contrib.sessions.middleware import SessionMiddleware
    from django.contrib.auth.middleware import AuthenticationMiddleware

except:
    class WSGIRequest:
        def __init__ (self, env):
            raise SystemError ("Django does not be installed or misconfigured")
    
else:
   DjangoSession = SessionMiddleware ()
   DjangoAuthentication = AuthenticationMiddleware ()


class Response:
    def __init__ (self, was):
        self.was = was
    
    @property
    def status_code (self):
        return int (self.was.response.get_status () [:3])
                    
    def set_cookie (self, k, v, max_age, expires, domain, path, secure, httponly):
        self.was.cookie.set (k, v, expires, path, domain, secure, httponly)        
    
    def delete_cookie (self, key, path, domain):
        self.was.cookie.remove (key, path, domain)
    
    def has_header (self, k):
        return self.was.response.get_header (k)
    
    def __getitme__ (self, k):
        return self.was.response [k]
        
    def __setitem__ (self, k, v):
        self.was.response [k] = v


class DjangoRequest (WSGIRequest):
    def __init__ (self, was):
        self.was = was
        WSGIRequest.__init__ (self, was.env)
        
        # making self.session
        DjangoSession.process_request (self)
        # making self.user
        DjangoAuthentication.process_request (self)        
        
    def authenticate (self, username, password):
        return auth.authenticate (self, username = username, password = password)
    
    def login (self, user):
        auth.login (self, user)
        
    def logout (self):
        auth.logout (self)
        
    def update_session_auth_hash (self, user):
        auth.update_session_auth_hash (self, user)
                
    def commit (self):
        response = Response (self.was)
        DjangoSession.process_response (self, response)


def request (was):
    if not WSGIRequest:
        raise SystemError ("Django not installed")
    request = DjangoRequest (was)    
    return request



from .secured_cookie_value import SecuredDictValue
import time

class Session (SecuredDictValue):
	default_session_timeout = 1200 # 20 min.	
	KEY = "sdlsess"
	VALIDS = "_valids"
	
	def __init__ (self, name, cookie, request, secret_key, session_timeout = 0):
		self.session_timeout = session_timeout and session_timeout or self.default_session_timeout
		self.new_session_timeout = None
		SecuredDictValue.__init__ (self, name, cookie, request, secret_key)
	
	def _recal_expires (self, expires):
		if expires is None:
			if self.new_session_timeout is not None:
				expires = self.new_session_timeout			
			else:
				return self.session_timeout
		if expires == "now":
			return 0
		if expires == "never":
			raise ValueError("session must be specified expires seconds")
		return int (expires)
			
	def validate (self):
		if self.VALIDS not in self.data:
			self.data = {}
			return
			
		if type (self.data [self.VALIDS]) is tuple:
			expires, addr = self.data [self.VALIDS]
			self._source_verified = (addr == self.request.get_remote_addr ())				
		else:
			expires = self.data [self.VALIDS]
			
		if time.time() > expires: # expired
			self.data = {}
			return
			
		if self.session_timeout + time.time () < expires: # too long, maybe fraud
			self.data = {}
			return
	
	def getv (self, k, v = None):
		if not self._source_verified:
			self.data = {}
			self.dirty = True
			return v
		return self.get (k, v)
	
	def touch (self):
		self.dirty = True
	
	def set_expiry (self, timeout):		
		self.new_session_timeout = timeout
		self.dirty = True
	
	def get_expiry (self):		
		if not self.data.get (self.VALIDS):
			return
		return self.data [self.VALIDS][0]
		
	def expire (self):
		self.clear ()
		self.new_session_timeout = 'now'
	
	def commit (self, expires = None):
		if not self.dirty or self.data is None: 
			return
		
		if len (self.data) == 1 and self.VALIDS in self.data: # only have _expires, expire now
			expires = 0
		else:	
			expires = self._recal_expires (expires)
		
		if not expires:
			self [self.VALIDS] = (time.time (), self.request.get_remote_addr ())
			
		else:	
			new = time.time () + expires
			if "_expire" not in self.data:
				self [self.VALIDS] = (new, self.request.get_remote_addr ())
			else:	
				current = self [self.VALIDS][0]				
				if self.new_session_timeout or new > current:
					#already set_expiry
					self [self.VALIDS] = (new, self.request.get_remote_addr ())
				else:
					expires = current - time.time ()
		
		self.set_cookie (expires)
		self.dirty = False
	


import os, sys
import time
import tempfile
from hmac import new as hmac
from hashlib import sha1
import json
import xmlrpc.client as xmlrpclib
import base64
import pickle
from multiprocessing import RLock
import random
import threading
from aquests.lib import pathtool, logger, jwt
from aquests.lib.producers import simple_producer, file_producer
from aquests.lib.athreads import trigger
from aquests.protocols.smtp import composer
from aquests.protocols.http import http_date
import inspect
from skitai import __version__, WS_EVT_OPEN, WS_EVT_CLOSE, WS_EVT_INIT
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB
from skitai import lifetime
from skitai.server.handlers import api_access_handler, vhost_handler
from .rpc import cluster_manager, cluster_dist_call
from .dbi import cluster_manager as dcluster_manager, cluster_dist_call as dcluster_dist_call
from . import server_info
from . import http_response
from .wastuff.promise import Promise, _Method
from .wastuff.triple_logger import Logger
from .wastuff import django_adaptor
from .wastuff.api import DateEncoder

if os.name == "nt":
	TEMP_DIR =  os.path.join (tempfile.gettempdir(), "skitai-gentemp")
else:
	TEMP_DIR = "/var/tmp/skitai-gentemp"		
pathtool.mkdir (TEMP_DIR)

class JWTUser:
	def __init__ (self, claims):
		self.__claims = claims
	
	@property
	def name (self):
		return self.__claims ["username"]
    	
	def __getattr__ (self, attr):
		return self.__claims.get (attr)
            
	def __str__ (self):
		return self.name


class WAS:
	version = __version__	
	objects = {}	
	_luwatcher = None
	
	lock = RLock ()
	init_time = time.time ()	
	
	# application friendly methods -----------------------------------------
	
	@classmethod
	def register (cls, name, obj):
		if hasattr (cls, name):
			raise AttributeError ("server object `%s` is already exists" % name)
		cls.objects [name] = obj
		setattr (cls, name, obj)
	
	@classmethod
	def unregister (cls, name):
		del cls.objects [name]
		return delattr (cls, name)
		
	@classmethod
	def add_handler (cls, back, handler, *args, **karg):
		h = handler (cls, *args, **karg)
		if hasattr (cls, "httpserver"):
			cls.httpserver.install_handler (h, back)
		return h
						
	@classmethod
	def add_cluster (cls, clustertype, clustername, clusterlist, ssl = 0, access = []):
		if clustertype and clustertype [0] == "*":
			clustertype = clustertype [1:]
		ssl = 0
		if ssl in (1, True, "1", "yes") or clustertype in ("https", "wss", "grpcs", "rpcs"):
			ssl = 1
		if type (clusterlist) is str:
			clusterlist = [clusterlist]	

		if clustertype and "*" + clustertype in (DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB):
			cluster = dcluster_manager.ClusterManager (clustername, clusterlist, "*" + clustertype, access, cls.logger.get ("server"))
			cls.clusters_for_distcall [clustername] = dcluster_dist_call.ClusterDistCallCreator (cluster, cls.logger.get ("server"))
		else:
			cluster = cluster_manager.ClusterManager (clustername, clusterlist, ssl, access, cls.logger.get ("server"))
			cls.clusters_for_distcall [clustername] = cluster_dist_call.ClusterDistCallCreator (cluster, cls.logger.get ("server"), cls.cachefs)
		cls.clusters [clustername] = cluster
	
	@property
	def timestamp (self):
		return int (time.time () * 1000) 
	
	@property
	def uniqid (self):
		return "{}{}".format (self.timestamp, self.gentemp () [-7:])
		
	def __dir__ (self):
		return self.objects.keys ()
	
	def __str__ (self):
		return "skitai was for {}".format (threading.currentThread ())
					
	def __detect_cluster (self, clustername):
		try: 
			clustername, uri = clustername.split ("/", 1)
		except ValueError:
			clustername, uri = clustername, ""
		if clustername [0] == "@":
			clustername = clustername [1:]
		
		try: 
			return self.clusters_for_distcall ["{}:{}".format (clustername, self.app.app_name)], "/" + uri			
		except KeyError:
			return self.clusters_for_distcall [clustername], "/" + uri
		
	def in__dict__ (self, name):
		return name in self.__dict__
	
	def _clone (self, disable_aquests = False):
		new_was = self.__class__ ()
		for k, v in self.__dict__.items ():
			setattr (new_was, k, v)	
		if disable_aquests:
			new_was.VALID_COMMANDS = []
		return new_was
	
	# mehiods remap ------------------------------------------------
			
	VALID_COMMANDS = [
		"options", "trace", "upload",
		"get", "getjson",
		"delete", "deletejson",
		"post", "postjson",
		"put", "putjson",
		"patch", "patchjson",
		"rpc", "jsonrpc", "grpc", "ws", "wss", 
		"db", "postgresql", "sqlite3", "redis", "mongodb", "backend",		
	]
	def __getattr__ (self, name):
		# method magic		
		if name in self.VALID_COMMANDS:
			return _Method (self._call, name, inspect.stack() [1])
		
		if self.in__dict__ ("app"): # saddle app			
			attr = self.app.create_on_demand (self, name)
			if attr:
				setattr (self, name, attr)
				return attr
		
		try:
			return self.objects [name]
		except KeyError:	
			raise AttributeError ("'was' hasn't attribute '%s'" % name)	
	
	def _call (self, method, args, karg):
		# was.db, 		was.get, 			was.post,			was.put, ...
		# was.db.lb, 	was.get.lb,		was.post.lb,	was.put.lb, ...
		# was.db.map,	was.get.map,	was.post.map,	was.put.map, ...

		uri = None
		if args:		uri = args [0]
		elif karg:	uri = karg.get ("uri", "")
		if not uri:	raise AssertionError ("Missing param uri or cluster name")

		try: 
			command, fn = method.split (".")
		except ValueError: 
			command = method
			if uri [0] == "@": 
				fn = "lb"
			else:
				fn = (command in ("db", "postgresql", "sqlite3", "redis", "mongodb", "backend") and "db" or "rest")

		if fn == "map" and not hasattr (self, "threads"):
			raise AttributeError ("Cannot use Map-Reduce with Single Thread")
		
		if command == "db":
			return getattr (self, "_d" + fn) (*args, **karg)			
		elif command in ("postgresql", "sqlite3", "redis", "mongodb", "backend"):
			return getattr (self, "_a" + fn) ("*" + command, *args, **karg)					
		else:	
			return getattr (self, "_" + fn) (command, *args, **karg)
	
	# TXN -----------------------------------------------
	
	def txnid (self):
		return "%s/%s" % (self.request.gtxid, self.request.ltxid)
	
	def rebuild_header (self, header = None):
		if not header:
			nheader = {}
		elif type (header) is list:
			nheader = {}			
			for k, v in header:
				nheader [k] = v
		else:
			nheader = header
					
		nheader ["X-Gtxn-Id"] = self.request.get_gtxid ()
		nheader ["X-Ltxn-Id"] = self.request.get_ltxid (1)
		return nheader
	
	# async requests -----------------------------------------------
		
	def _rest (self, method, uri, data = None, auth = None, headers = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
		return self.clusters_for_distcall ["__socketpool__"].Server (uri, data, method, self.rebuild_header (headers), auth, meta, use_cache, False, filter, callback, timeout, caller)
	
	def _crest (self, mapreduce = False, method = None, uri = None, data = None, auth = None, headers = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
		cluster, uri = self.__detect_cluster (uri)
		return cluster.Server (uri, data, method, self.rebuild_header (headers), auth, meta, use_cache, mapreduce, filter, callback, timeout, caller)
				
	def _lb (self, *args, **karg):
		return self._crest (False, *args, **karg)
		
	def _map (self, *args, **karg):
		return self._crest (True, *args, **karg)
	
	def _bind_sqlphile (self, dbo):
		try:
			app_sqlphile = self.sql
		except AttributeError:
			return dbo	
		return app_sqlphile.new (dbo)
			
	def _ddb (self, server, dbname = "", auth = None, dbtype = DB_PGSQL, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
		dbo = self.clusters_for_distcall ["__dbpool__"].Server (server, dbname, auth, dbtype, meta, use_cache, False, filter, callback, timeout, caller)
		if dbtype in (DB_PGSQL, DB_SQLITE3):
			return self._bind_sqlphile (dbo)
		return dbo
	
	def _cddb (self, mapreduce = False, clustername = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
		if mapreduce and callback: raise RuntimeError ("Cannot use callback with Map-Reduce")
		cluster = self.__detect_cluster (clustername) [0]
		dbo = cluster.Server (None, None, None, None, meta, use_cache, mapreduce, filter, callback, timeout, caller)
		if cluster.cluster.dbtype in (DB_PGSQL, DB_SQLITE3):
			return self._bind_sqlphile (dbo)
		return dbo	
	
	def _dlb (self, *args, **karg):
		return self._cddb (False, *args, **karg)
	
	def _dmap (self, *args, **karg):
		return self._cddb (True, *args, **karg)
	
	def _adb (self, dbtype, server, dbname = "", auth = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
		return self._ddb (server, dbname, auth, dbtype, meta, use_cache, filter, callback, timeout, caller)
	
	def _alb (self, dbtype, *args, **karg):
		return self._cddb (False, *args, **karg)
	
	def _amap (self, dbtype, *args, **karg):
		return self._cddb (True, *args, **karg)
	
	# system functions ----------------------------------------------
		
	def log (self, msg, category = "info", at = "app"):
		self.logger (at, msg, "%s:%s" % (category, self.txnid ()))
		
	def traceback (self, id = "", at = "app"):
		if not id:
			id = self.txnid ()
		self.logger.trace (at, id)
	
	@property
	def tempfile (self):
		return self.gentemp () 
	
	def gentemp (self):
		return os.path.join (TEMP_DIR, next (tempfile._get_candidate_names()))
	
	# -- only allpy to current worker process
	def status (self, flt = None, fancy = True):		
		return server_info.make (self, flt, fancy)
	
	def restart (self, timeout = 0):
		lifetime.shutdown (3, timeout)
	
	def shutdown (self, timeout = 0):
		lifetime.shutdown (0, timeout)
	
	# URL builders -------------------------------------------------
	def urlfor (self, thing, *args, **karg):
		# override with resource default args
		if not isinstance (thing, str) or thing.startswith ("/") or thing.find (":") == -1:
			return self.app.urlfor (thing, *args, **karg)		
		return self.apps.urlfor (thing, *args, **karg)	
	ab = urlfor
	
	def partial (self, thing, **karg):
		# override with current args
		karg ["__defaults__"] = self.request.args
		return self.ab (thing, **karg)
	
	def baseurl (self, thing):
		# resource path info without parameters
		return self.ab (thing, __resource_path_only__ = True)
	basepath = baseurl
	
	# response helpers --------------------------------------------
		
	def render (self, template_file, _do_not_use_this_variable_name_ = {}, **karg):
		return self.app.render (self, template_file, _do_not_use_this_variable_name_, **karg)
	
	REDIRECT_TEMPLATE =  (
		"<html><head><title>%s</title></head>"
		"<body><h1>%s</h1>"
		"This document may be found " 
		'<a HREF="%s">here</a></body></html>'
	)
	def redirect (self, url, status = "302 Object Moved", body = None, headers = None):
		redirect_headers = [
			("Location", url), 
			("Cache-Control", "max-age=0"), 
			("Expires", http_date.build_http_date (time.time ()))
		]
		if type (headers) is list:
			redirect_headers += headers
		if not body:
			body = self.REDIRECT_TEMPLATE % (status, status, url)			
		return self.response (status, body, redirect_headers)
	
	def promise (self, handler, **karg):
		self.response.set_streaming ()
		return Promise (self, handler, **karg)
	
	def email (self, subject, snd, rcpt):
		if composer.Composer.SAVE_PATH is None:
			composer.Composer.SAVE_PATH = os.path.join ("/var/tmp/skitai/default", "smtpda", "mail", "spool")
			pathtool.mkdir (composer.Composer.SAVE_PATH)
		return composer.Composer (subject, snd, rcpt)
	
	# event -------------------------------------------------
	
	def broadcast (self, event, *args, **kargs):
		return self.apps.bus.emit (event, self, *args, **kargs)
	
	def setlu (self, name, *args, **karg):
		self._luwatcher.set (name, time.time (), karg.get ("x_ignore", False))
		self.broadcast (name, *args, **karg)			
		
	def getlu (self, *names):
		mtimes = []
		for name in names:
			mtime = self._luwatcher.get (name, self.init_time)
			mtimes.append (mtime)
		return max (mtimes)
	
	# JWT token --------------------------------------------------
	def mkjwt (self, claim, alg = "HS256"):
		return jwt.gen_token (self.app.salt, claim, alg)
	
	def dejwt (self, token):		
		try: 
			claims = jwt.get_claim (self.app.salt, token)
		except (TypeError, ValueError): 
			return	
		if claims is None:
			return
		if "username" in claims:
			self.request.user = JWTUser (claims)
		return claims
		
	# simple/session token  ----------------------------------------
	def _unserialize (self, string):
		def adjust_padding (s):
			paddings = 4 - (len (s) % 4)
			if paddings != 4:
				s += ("=" * paddings)
			return s
		
		string = string.replace (" ", "+")
		try:
			base64_hash, data = string.split('?', 1)
		except ValueError:
			return	
		client_hash = base64.b64decode(adjust_padding (base64_hash))
		data = base64.b64decode(adjust_padding (data))
		mac = hmac (self.app.salt, None, sha1)		
		mac.update (data)
		if client_hash != mac.digest():
			return
		return pickle.loads (data)
	
	def mktoken (self, obj, timeout = 1200, session_key = None):
		wrapper = {
			'object': obj,
			'timeout': time.time () + timeout
		}
		if session_key:
			token = hex (random.getrandbits (64))
			tokey = '_{}_token'.format (session_key)
			wrapper ['_session_token'] = (tokey, token)
			self.session [tokey] = token
			
		data = pickle.dumps (wrapper, 1)
		mac = hmac (self.app.salt, None, sha1)
		mac.update (data)
		return (base64.b64encode (mac.digest()).strip().rstrip (b'=') + b"?" + base64.b64encode (data).strip ().rstrip (b'=')).decode ("utf8")
	token = mktoken
	
	def detoken (self, string):
		wrapper = self._unserialize (string)
		if not wrapper:
			return 
		
		# validation with session
		tokey = None
		has_error = False
		if wrapper ['timeout']  < time.time ():
			has_error = True

		if not has_error:
			session_token = wrapper.get ('_session_token')
			if session_token:
				# verify with session				
				tokey, token = session_token	
				if token != self.session.get (tokey):
					has_error = True
					
		if has_error:
			if tokey:
				del self.session [tokey]
			return
		
		obj = wrapper ['object']
		return obj
	
	def rmtoken (self, string):
		wrapper = self._unserialize (string)
		session_token = wrapper.get ('_session_token')
		if not session_token:
			return
		tokey, token = session_token
		if not self.session.get (tokey):
			return
		del self.session [tokey]
		
	# CSRF token ------------------------------------------------------	
	CSRF_NAME = "_csrf_token"
	@property
	def csrf_token (self):
		if self.CSRF_NAME not in self.session:
			self.session [self.CSRF_NAME] = hex (random.getrandbits (64))
		return self.session [self.CSRF_NAME]

	@property
	def csrf_token_input (self):
		return '<input type="hidden" name="{}" value="{}">'.format (self.CSRF_NAME, self.csrf_token)
	
	def csrf_verify (self, keep = False):
		if not self.request.args.get (self.CSRF_NAME):
			return False
		token = self.request.args [self.CSRF_NAME]
		if self.csrf_token == token:
			if not keep:
				del self.session [self.CSRF_NAME]
			return True
		return False
	
	# proxy & adaptor  -----------------------------------------------
	@property
	def sql (self):
		return self.app.sqlphile
	
	@property
	def django (self):
		if hasattr (self.request, "django"):
			return self.request.django
		self.request.django = django_adaptor.request (self)
		return self.request.django
	
	# websocket methods for generic WSGI containers -----------------------------
	
	def wsconfig (self, spec, timeout = 60, encoding = "text"):
		self.env ["websocket.config"] = (spec, timeout, encoding)
		return ""
		
	def wsinit (self):
		return self.env.get ('websocket.event') == WS_EVT_INIT
	
	def wsopened (self):
		return self.env.get ('websocket.event') == WS_EVT_OPEN
	
	def wsclosed (self):
		return self.env.get ('websocket.event') == WS_EVT_CLOSE
	
	def wshasevent (self):
		return self.env.get ('websocket.event')
	
	def wsclient (self):
		return self.env.get ('websocket.client')	
			
	# will be deprecated --------------------------------------------------	
	
	def togrpc (self, obj):
		return obj.SerializeToString ()
	
	def fromgrpc (self, message, obj):
		return message.ParseFromString (obj)
		
	def tojson (self, obj):
		return json.dumps (obj, cls = DateEncoder)
		
	def toxml (self, obj):
		return xmlrpclib.dumps (obj, methodresponse = False, allow_none = True, encoding = "utf8")	
	
	def fromjson (self, obj):
		if type (obj) is bytes:
			obj = obj.decode ('utf8')
		return json.loads (obj)
	
	def fromxml (self, obj, use_datetime = 0):
		return xmlrpclib.loads (obj)
	
	def fstream (self, path, mimetype = 'application/octet-stream'):	
		self.response.set_header ('Content-Type',  mimetype)
		self.response.set_header ('Content-Length', str (os.path.getsize (path)))	
		return file_producer (open (path, "rb"))
			
	def jstream (self, obj, key = None):
		self.response.set_header ("Content-Type", "application/json")
		if key:
			# for single skeleton data is not dict
			return self.tojson ({key: obj})
		else:
			return self.tojson (obj)		
	
	def xstream (self, obj, use_datetime = 0):			
		self.response.set_header ("Content-Type", "text/xml")
		return self.toxml (obj, use_datetime)
	
	def gstream (self, obj):
		self.response.set_header ("Content-Type", "application/grpc")
		return self.togrpc (obj)
	
	def render_ei (self, exc_info, format = 0):
		return http_response.catch (format, exc_info)
		


import asyncore
import time

def _make (was, flt):
	g = {}
	
	if flt:
		if flt.startswith ("CLUSTER:"):
			return {flt: was.clusters [flt[8:]].status ()}
		elif flt == "SOCKETPOOL":
			return {flt: was.clusters [None].status ()}
		elif flt == "ENVIRON":
			return {"ENVIRON": was.env}	
	else:
		for name, cluster in list(was.clusters.items ()):
			if name is None: 
				g ["SOCKETPOOL"] = cluster.status ()
				continue
			g ["CLUSTER:" + name] = cluster.status ()
			g ["ENVIRON"] = was.env
			
	for attr in dir (was):
		if attr.startswith ("__") or attr == "config": continue
		if flt and flt != attr.upper (): continue
		obj = eval ("was." + attr)
		if hasattr (obj, "status"):			
			g [attr.upper ()] = obj.status ()
	
	return g
	
	
def format_object (o, depth = 0):
	b = []
	if type (o) in (type ([]), type (())):
		for each in o:
			if type (each) is type ({}):
				b.append ("%s" % format_object (each, depth + 1))
			elif type (each) is type ([]):
				b.append (", ".join ([str (x) for x in each]))				
			else:					
				if not b:
					b.append ("<ul>")
				b.append ("<li>%s</li>" % str (each))
		if b and b [0] == "<ul>":
			b.append ("</ul>")		
		return "".join (b)

	b.append ("<table width='100%' border='1'>")
	ii = list(o.items ())
	ii.sort ()
	width = 15 + (depth * 6)
	for k1, v1 in ii:
		if type (v1) in (type ([]), type ({})):
			b.append ("<tr><td bgcolor='#efefef' valign='top' width='%d%%'><b>%s%s</b></td>" % (width, str (k1), type (v1) in (list, tuple, dict) and " (%d)" % (len (v1),) or ""))			
			b.append ("<td valign='top'>%s</td></tr>" % format_object (v1, depth + 1))			
		else:
			b.append ("<tr><td bgcolor='#efefef' valign='top' width='%d%%'>%s</td>" % (width, str (k1)))
			b.append ("<td valign='top'>%s</td></tr>" % str (v1))
	b.append ("</table>")	
	return "".join (b)

	
def formatting (was, info, flt):	
	b = []
	l = list(info.items ())
	l.sort ()
	if not flt:
		b.append ('<a name="TOC"></a><h3>Table of Content</h3>')
		b.append ('<ol>')
		for k, v in l:
			b.append ("<li><a href='#%s'>%s</a></li>" % (k, k))		
		b.append ('</ol>')
	
	for k, v in l:
		if not flt:
			b.append ("<a name='%s'></a><h3><a href='%s?f=%s'>%s</a></h3>" % (k, was.request.uri, k, k))
		else:
			b.append ("<h3><a href='%s'>%s</a></h3>" % (was.request.split_uri ()[0], k))	
			
		b.append (format_object (v))
		if not flt:
			b.append ('<div style="margin-bottom: 30px; padding: 5px;"><a href="#TOC">TOC</a></div>')
		else:
			b.append ('<div style="margin-bottom: 30px; padding: 5px;"><a href="%s">TOC</a></div>' % was.request.split_uri ()[0])
	return "".join (b)
	

def make (was, flt = None, fancy = True):
	info = _make (was, flt)
	if fancy:
		return formatting (was, info, flt)	
	return info




import tempfile
import os
try:
	from StringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO	

class FormCollector:
	def __init__ (self, handler, request, *args):
		self.handler = handler
		self.request = request
		self.buffer = BytesIO ()
		self.content_length = self.get_content_length ()
		self.size = 0
		self.max_cl = 0 
	
	def set_max_content_length (self, max_cl):
		self.max_cl = max_cl
	
	def get_max_content_length (self):
		return self.max_cl
			
	def get_content_length (self):
		cl = self.request.get_header ('content-length')
		if cl is not None:
			try:
				cl = int  (cl)
			except (ValueError, TypeError):
				cl = None
		return cl
		
	def start_collect (self):
		if self.content_length == 0: 
			return self.found_terminator()
		self.request.channel.set_terminator (self.content_length)

	def collect_incoming_data (self, data):
		self.size += len (data)
		if self.max_cl and self.size > self.max_cl:
			raise ValueError ("Too Large Entity")
		self.buffer.write (data)

	def found_terminator (self):
		# prepare got recving next request header
		self.request.collector = None  # break circ. ref
		self.request.channel.set_terminator (b'\r\n\r\n')
		
		body = self.buffer.getvalue ()
		if body:
			self.request.set_body (body)
			self.buffer.seek (0)
			self.handler.continue_request (self.request, self.buffer)
		else:
			self.handler.continue_request (self.request, None)	
	
	def close (self):
		self.buffer.close ()
		self.request.collector = None  # break circ. ref


class HTTP2DummyCollector (FormCollector):
	def __init__ (self, handler, request, respcode):
		self.handler = handler
		self.request = request		
		self.respcode = respcode
		self.content_length = self.get_content_length ()
		self.size = 0
		self.max_cl = 1048576 # allow max 1M
		if self.content_length and self.content_length > self.max_cl:
			self.finish_collect (True)
	
	def finish_collect (self, force_close = False):
		self.request.channel.set_terminator (9)
		self.request.collector = None  # break circ. ref
		self.handler.continue_request (self.request, None, (self.respcode, force_close))
		
	def collect_incoming_data (self, data):
		self.size += len (data)
		if self.max_cl and self.size > self.max_cl:
			self.finish_collect (True)

	def found_terminator (self):
		self.finish_collect ()
	
	def close (self):		
		self.request.collector = None  # break circ. ref


class MultipartCollector (FormCollector):
	def __init__ (self, handler, request, upload_max_size, file_max_size, cache_max_size, *args):
		self.handler = handler
		self.request = request		
		self.content_length = self.get_content_length ()
		self.buffer = tempfile.NamedTemporaryFile(delete=False)
		self.size = 0
								
	def start_collect (self):
		if self.content_length == 0: 
			return self.found_terminator()		
		self.request.channel.set_terminator (self.content_length)
	
	def collect_incoming_data (self, data):
		self.size += len (data)
		self.buffer.write (data)
		
	def close (self):
		self.buffer.close ()
		self.request.collector = None
				
	def found_terminator (self):
		self.buffer.close ()		
		self.handler.continue_request (self.request, open (self.buffer.name, "rb"))
		self.request.channel.set_terminator (b'\r\n\r\n')



"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""
import re
import sys
import os
import shutil, glob
import codecs
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

with open('skitai/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)

if 'develop' in sys.argv[1:]:
	# For reseting entry point
	os.system ("rm -rf /usr/local/lib/python3.5/dist-packages/aquests-*")
	
if sys.argv[-1] == 'publish':
	buildopt = ['sdist', 'upload']	
	if os.name == "nt":
		buildopt.insert (0, 'bdist_wheel')
	os.system('python setup.py %s' % " ".join (buildopt))
	for each in os.listdir ("dist"):
		os.remove (os.path.join ('dist', each))
	sys.exit()
	
classifiers = [
  'License :: OSI Approved :: MIT License',
  'Development Status :: 4 - Beta',
  'Topic :: Internet :: WWW/HTTP :: HTTP Servers',	
	'Topic :: Internet :: WWW/HTTP :: WSGI',
	'Environment :: Console',
	'Environment :: No Input/Output (Daemon)',
	'Topic :: Internet',
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',
	'Programming Language :: Python',	
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.4',
	'Programming Language :: Python :: 3.5',
]

packages = [
	'skitai',
	'skitai.bin',
	'skitai.server',
	'skitai.server.dbi',
	'skitai.server.wastuff',
	'skitai.server.handlers',
	'skitai.server.rpc',	
	'skitai.server.offline',
	'skitai.server.handlers.http2',
	'skitai.server.handlers.websocket',
	'skitai.server.handlers.proxy',
	'skitai.saddle',
	'skitai.saddle.patches',
	'skitai.saddle.contrib',
	'skitai.saddle.contrib.decorative',	
	'skitai.saddle.contrib.decorative.django',
]

package_dir = {'skitai': 'skitai'}
package_data = {
	"skitai": [
		"saddle/contrib/templates/*"
	]
}

install_requires = [
	"jinja2==2.10",
	"aquests==0.7.18",
	"sqlphile==0.4.7",
	"jsonrpclib-pelix==0.3.1",
	"mock==2.0.0"
]

with codecs.open ('README.rst', 'r', encoding='utf-8') as f:
	long_description = f.read()
    
setup (
	name='skitai',
	version=version,
	description='Skitai App Engine',
	long_description=long_description,
	url = 'https://gitlab.com/hansroh/skitai',
	author='Hans Roh',
	author_email='hansroh@gmail.com',	
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	entry_points = {
        'console_scripts': [
			'skitai-smtpda=skitai.bin.smtpda:main',
			'skitai-cron=skitai.bin.cron:main'
		],
    },
	license='MIT',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/skitai",
	install_requires = install_requires,
	classifiers=classifiers	
)



from aquests.lib import producers
import threading
import os

class SelectiveLogger:
	MAX_CACHE = 1000
	
	def __init__ (self, log_off = []):
		self.log_off = log_off
		self.endswiths = []
		self.startswiths = []
		self.cache = {}
		self.lock = threading.RLock ()
		for each in log_off:
			if each.startswith ("*"):
				self.endswiths.append (each [1:])
			elif each.endswith ("*"):
				self.startswiths.append (each[:-1])
			else:				
				self.startswiths.append (each)		
	
	def add_cache (self, key, val):	
		with self.lock:
			if len (self.cache) > self.MAX_CACHE:
				return
			self.cache [key] = val
		
	def __call__ (self, uri, producer, logfunc):
		if not self.log_off or uri == "/":
			return producers.hooked_producer (producer, logfunc)		
		
		with self.lock:
			cached = self.cache.get (uri)
		if cached == -1:
			return producer
		
		for each in self.endswiths:
			if uri.endswith (each):
				self.add_cache (uri, -1)
				return producer
				
		d = os.path.dirname (uri)
		with self.lock:
			cached = self.cache.get (d)
		if cached == -1:
			return producer
		elif cached == 1:
			return producers.hooked_producer (producer, logfunc)
		
		for each in self.startswiths:
			if uri.startswith (each):
				self.add_cache (d, -1)
				return producer

		self.add_cache (d, 1)
		return producers.hooked_producer (producer, logfunc)		
		